/*
 *	MP3 bitstream Output interface for LAME
 *
 *	Copyright (c) 1999 Takehiro TOMINAGA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#include "config.h"
#include "global.h"
#include "lame.h"
#include "util.h"
#include "tables.h"
#include "bitstream.h"
#include "quantize.h"
#include "quantize_pvt.h"
#include "gogo_io.h"

/* This is the scfsi_band table from 2.4.2.7 of the IS */
const int scfsi_band[5] = { 0, 6, 11, 16, 21 };

/* unsigned int is at least this large:  */
/* we work with ints, so when doing bit manipulation, we limit
 * ourselves to MAX_LENGTH-2 just to be on the safe side */
#define MAX_LENGTH      32

static void putheader_bits(void)
{
	Bit_stream_struc *bs;
	bs = &RW.bs;
	memcpy(&bs->buf[bs->buf_byte_idx], RW.header[RW.w_ptr].buf, RO.sideinfo_len);
	bs->buf_byte_idx += RO.sideinfo_len;
	bs->totbit += RO.sideinfo_len * 8;
	RW.w_ptr = (RW.w_ptr + 1) & (MAX_HEADER_BUF - 1);
}

/*write j bits into the bit stream */
static void putbits2(int val, int j)
{
	Bit_stream_struc *const bs = &RW.bs;
	assert(j < MAX_LENGTH - 2);

	/* 0 <= buf_bit_idx <= 8 */
	while (j > 0) {
		int k;
		if (bs->buf_bit_idx == 0) {
			bs->buf_bit_idx = 8;
			bs->buf_byte_idx++;
			assert(bs->buf_byte_idx < LAME_MAXMP3BUFFER);
			assert(RW.header[RW.w_ptr].write_timing >= bs->totbit);
			if (RW.header[RW.w_ptr].write_timing == bs->totbit) {
				putheader_bits();
			}
			bs->buf[bs->buf_byte_idx] = 0;
		}
		k = Min(j, bs->buf_bit_idx);
		j -= k;
		bs->buf_bit_idx -= k;
		assert(bs->buf_bit_idx < MAX_LENGTH);
		bs->buf[bs->buf_byte_idx] |= ((val >> j) << bs->buf_bit_idx);
		bs->totbit += k;
	}
}

/*write j bits into the bit stream, ignoring frame headers */
static void putbits_noheaders(int val, int j)
{
	Bit_stream_struc *const bs = &RW.bs;
	assert(j < MAX_LENGTH - 2);

	while (j > 0) {
		int k;
		if (bs->buf_bit_idx == 0) {
			bs->buf_bit_idx = 8;
			bs->buf_byte_idx++;
			assert(bs->buf_byte_idx < LAME_MAXMP3BUFFER);
			bs->buf[bs->buf_byte_idx] = 0;
		}

		k = Min(j, bs->buf_bit_idx);
		j -= k;
		bs->buf_bit_idx -= k;
		assert(bs->buf_bit_idx < MAX_LENGTH);
		bs->buf[bs->buf_byte_idx] |= ((val >> j) << bs->buf_bit_idx);
		bs->totbit += k;
	}
}


/*
  Some combinations of bitrate, Fs, and stereo make it impossible to stuff
  out a frame using just main_data, due to the limited number of bits to
  indicate main_data_length. In these situations, we put stuffing bits into
  the ancillary data...
*/

static void drain_into_ancillary(int remainingBits)
{
	assert(remainingBits >= 0);

	while (remainingBits >= 16) {
#if defined(CPU_I386)
		putbits2(0x474F, 16);
		remainingBits -= 16;
#else
		putbits2(0x47, 8);
		remainingBits -= 8;
		putbits2(0x4F, 8);
		remainingBits -= 8;
#endif
	}
#if defined(CPU_I386)
	if (remainingBits == 0)
		return;
	{
		int i;
		i = ((RW.ancillary_flag == (remainingBits & 1)) ? 0x5555 : 0xAAAA) & ((1 << remainingBits) - 1);
		if (remainingBits & 1)
			RW.ancillary_flag ^= 1;
		putbits2(i, remainingBits);
	}
#else
	for (; remainingBits >= 1; remainingBits -= 1) {
		putbits2(RW.ancillary_flag, 1);
		RW.ancillary_flag ^= 1;
	}
	assert(remainingBits == 0);
#endif
}

/*write N bits into the header */
static void writeheader(int val, int j)
{
	int ptr = RW.header[RW.h_ptr].ptr;
	while (j > 0) {
		int k = Min(j, 8 - (ptr & 7));
		j -= k;
		assert(j < MAX_LENGTH);	/* >> 32  too large for 32 bit machines */
		RW.header[RW.h_ptr].buf[ptr >> 3] |= ((val >> j)) << (8 - (ptr & 7) - k);
		ptr += k;
	}
	RW.header[RW.h_ptr].ptr = ptr;
}


/* (jo) this wrapper function for BF_addEntry() updates also the crc */

/* 10Kclk@PIII */
/*  CRCʬ 6Kclk */
/* ޤȤ 3.3Kclk */
static void encodeSideInfo2(gogo_thread_data * tl, int bitsPerFrame)
{
	III_side_info_t *l3_side;
	int gr, ch;
	register unsigned int data;
	l3_side = &tl->l3_side;
	RW.header[RW.h_ptr].ptr = 0;
	memset(RW.header[RW.h_ptr].buf, 0, RO.sideinfo_len);

	data = (RO.out_samplerate < 16000) ? 0xFFE : 0xFFF;
	writeheader(data, 12);
	data = RO.version; data <<= 2;
	data |= (4 - 3 /* layer3 */ ); data <<= 1;
	data |= 1; /* !error_protection */ data <<= 4;
	data |= tl->bitrate_index; data <<= 2;
	data |= RO.samplerate_index; data <<= 1;
	data |= tl->padding; data <<= 1;
	/* 0 extension */ data <<= 2;
	data |= RO.mode; data <<= 2;
	data |= tl->mode_ext; data <<= 1;
	data |= COPYRIGHT; data <<= 1;
	data |= ORIGINAL; data <<= 2;
	data |= RO.emphasis;

	writeheader(data, 1 + 2 + 1 + 4 + 2 + 1 + 1 + 2 + 2 + 1 + 1 + 2);

	if (RO.version == 1) {
		/* MPEG1 */
		assert(RW.main_data_begin >= 0);
		writeheader(RW.main_data_begin, 9);

		if (RO.channels_out == 2) {
			/* 3bitl3_side->private_bits(=0) */
			data = l3_side->scfsi[0][0]; data <<= 1;
			data |= l3_side->scfsi[0][1]; data <<= 1;
			data |= l3_side->scfsi[0][2]; data <<= 1;
			data |= l3_side->scfsi[0][3]; data <<= 1;
			data |= l3_side->scfsi[1][0]; data <<= 1;
			data |= l3_side->scfsi[1][1]; data <<= 1;
			data |= l3_side->scfsi[1][2]; data <<= 1;
			data |= l3_side->scfsi[1][3];
			writeheader(data, 3 + 8);
		} else {
			/* 5bitl3_side->private_bits(=0) */
			data = l3_side->scfsi[0][0]; data <<= 1;
			data |= l3_side->scfsi[0][1]; data <<= 1;
			data |= l3_side->scfsi[0][2]; data <<= 1;
			data |= l3_side->scfsi[0][3]; 
			writeheader(data, 5 + 4);
		}

		for (gr = 0; gr < 2; gr++) {
			for (ch = 0; ch < RO.channels_out; ch++) {
				gr_info *gi = &tl->l3_side.tt[gr][ch];
				writeheader(gi->part2_3_length, 12);
				writeheader(gi->big_values / 2, 9);
				writeheader(gi->global_gain, 8);
				writeheader(gi->scalefac_compress, 4);
				writeheader(gi->window_switching_flag, 1);

				if (gi->table_select[0] == 14)
					gi->table_select[0] = 16;
				if (gi->table_select[1] == 14)
					gi->table_select[1] = 16;
				if (gi->window_switching_flag) {
					data = gi->block_type;
					data <<= 1;	data <<= 5;/* gi->mixed_block_flag = 0 */
					data |= gi->table_select[0]; data <<= 5;
					data |= gi->table_select[1]; data <<= 3;
					data |= gi->subblock_gain[0]; data <<= 3;
					data |= gi->subblock_gain[1]; data <<= 3;
					data |= gi->subblock_gain[2];
					writeheader(data, 2 + 1 + 5 + 5 + 3 + 3 + 3);
				} else {
					assert(gi->block_type == NORM_TYPE);
					assert(gi->region0_count < 16U && gi->region1_count < 8U);

					if (gi->table_select[2] == 14)
						gi->table_select[2] = 16;

					data = gi->table_select[0]; data <<= 5;
					data |= gi->table_select[1]; data <<= 5;
					data |= gi->table_select[2]; data <<= 4;
					data |= gi->region0_count; data <<= 3;
					data |= gi->region1_count;
					writeheader(data, 5 + 5 + 5 + 4 + 3);
				}
				data = gi->preflag; data <<= 1;
				data |= gi->scalefac_scale; data <<= 1;
				data |= gi->count1table_select;
				writeheader(data, 1 + 1 + 1);
			}
		}
	} else {
		/* MPEG2 */
		assert(RW.main_data_begin >= 0);
		writeheader((RW.main_data_begin), 8);
		writeheader(0 /* l3_side->private_bits */ , RO.channels_out);

		gr = 0;
		for (ch = 0; ch < RO.channels_out; ch++) {
			gr_info *gi = &tl->l3_side.tt[gr][ch];

			writeheader(gi->part2_3_length, 12);
			writeheader(gi->big_values / 2, 9);
			writeheader(gi->global_gain, 8);
			writeheader(gi->scalefac_compress, 9);
			writeheader(gi->window_switching_flag, 1);

			if (gi->window_switching_flag) {
				writeheader(gi->block_type, 2);
				writeheader(0 /*gi->mixed_block_flag */ , 1);

				if (gi->table_select[0] == 14)
					gi->table_select[0] = 16;
				writeheader(gi->table_select[0], 5);
				if (gi->table_select[1] == 14)
					gi->table_select[1] = 16;
				writeheader(gi->table_select[1], 5);

				writeheader(gi->subblock_gain[0], 3);
				writeheader(gi->subblock_gain[1], 3);
				writeheader(gi->subblock_gain[2], 3);
			} else {
				if (gi->table_select[0] == 14)
					gi->table_select[0] = 16;
				writeheader(gi->table_select[0], 5);
				if (gi->table_select[1] == 14)
					gi->table_select[1] = 16;
				writeheader(gi->table_select[1], 5);
				if (gi->table_select[2] == 14)
					gi->table_select[2] = 16;
				writeheader(gi->table_select[2], 5);

				assert(gi->region0_count < 16U);
				assert(gi->region1_count < 8U);
				writeheader(gi->region0_count, 4);
				writeheader(gi->region1_count, 3);
			}

			writeheader(gi->scalefac_scale, 1);
			writeheader(gi->count1table_select, 1);
		}
	}

	{
		int old = RW.h_ptr;
		assert(RW.header[old].ptr == RO.sideinfo_len * 8);

		RW.h_ptr = (old + 1) & (MAX_HEADER_BUF - 1);
		RW.header[RW.h_ptr].write_timing = RW.header[old].write_timing + bitsPerFrame;

		assert(RW.h_ptr != RW.w_ptr);	/* MAX_HEADER_BUF is too small. */
	}
}								/* encodeSideInfo2 */

static int huffman_coder_count1(int *ix, gr_info * gi, uint32 * xr_sign)
{
	/* Write count1 area */
	const struct huffcodetab *h = &ht[gi->count1table_select + 32];
	int i, bits = 0;

	ix += gi->big_values;
	xr_sign += gi->big_values;
	assert(gi->count1table_select < 2);

	for (i = (gi->count1 - gi->big_values) / 4; i > 0; --i) {
		int huffbits = 0;
		int p = 0;
		if (ix[0]) {
			p += 8;
			huffbits -= xr_sign[0];
		}
		if (ix[1]) {
			p += 4;
			huffbits = huffbits * 2 - xr_sign[1];
		}
		if (ix[2]) {
			p += 2;
			huffbits = huffbits * 2 - xr_sign[2];
		}
		if (ix[3]) {
			p += 1;
			huffbits = huffbits * 2 - xr_sign[3];
		}
		xr_sign += 4;
		ix += 4;
		putbits2(huffbits + h->table[p], h->hlen[p]);
		bits += h->hlen[p];
	}
	return bits;
}

/* Implements the pseudocode of page 98 of the IS */
/* NEED asm */
INLINE static int HuffmanCode(const int table_select, int x1, int x2, uint32 s1, uint32 s2)
{
	const struct huffcodetab *h = ht + table_select;
	int code = 0;
	int cbits = 0;
	int xbits = 0;
	int sgn_x1 = 0;
	int sgn_x2 = 0;
	int linbits = h->xlen;
	int xlen = h->xlen;
	int ext;

	assert(table_select > 0);

	/* x1 == 0 && s1 == -1 Ǥ뤳ȤΤ */
	if (x1 && s1) {
		sgn_x1++;
	}
	if (x2 && s2) {
		sgn_x2++;
	}

	ext = sgn_x1;

	if (table_select > 15) {
		/* use ESC-words */
		if (x1 > 14) {
			int linbits_x1 = x1 - 15;
			assert(linbits_x1 <= h->linmax);
			ext |= linbits_x1 << 1;
			xbits = linbits;
			x1 = 15;
		}

		if (x2 > 14) {
			int linbits_x2 = x2 - 15;
			assert(linbits_x2 <= h->linmax);
			ext <<= linbits;
			ext |= linbits_x2;
			xbits += linbits;
			x2 = 15;
		}
		xlen = 16;
	}

	if (x1 != 0) {
		cbits--;
	}

	if (x2 != 0) {
		ext <<= 1;
		ext |= sgn_x2;
		cbits--;
	}

	xbits -= cbits;

	assert((x1 | x2) < 16u);

	x1 = x1 * xlen + x2;

	code = h->table[x1];
	cbits += h->hlen[x1];

	assert(cbits <= MAX_LENGTH);
	assert(xbits <= MAX_LENGTH);

	putbits2(code, cbits);
	putbits2(ext, xbits);

	return cbits + xbits;
}

static int Huffmancodebits(int tableindex, int start, int end, int *ix, uint32 * xr_sign)
{
	int i, bits;

	assert(tableindex < 32);
	if (!tableindex)
		return 0;

	bits = 0;
	for (i = start; i < end; i += 2) {
		bits += HuffmanCode(tableindex, ix[i], ix[i + 1], xr_sign[i], xr_sign[i + 1]);
	}
	return bits;
}

/*
  Note the discussion of huffmancodebits() on pages 28
  and 29 of the IS, as well as the definitions of the side
  information on pages 26 and 27.
  */
static int ShortHuffmancodebits(int *ix, gr_info * gi, uint32 * xr_sign)
{
	int bits;
	int region1Start;

	region1Start = 3 * RO.scalefac_band.s[3];
	if (region1Start > gi->big_values)
		region1Start = gi->big_values;

	/* short blocks do not have a region2 */
	bits = Huffmancodebits(gi->table_select[0], 0, region1Start, ix, xr_sign);
	bits += Huffmancodebits(gi->table_select[1], region1Start, gi->big_values, ix, xr_sign);
	return bits;
}

static int LongHuffmancodebits(int *ix, gr_info * gi, uint32 * xr_sign)
{
	int i, bigvalues, bits = 0;
	int region1Start, region2Start;

	bigvalues = gi->big_values;
	assert(0 <= bigvalues && bigvalues <= 576);

	i = gi->region0_count + 1;
	assert(i < 23);
	region1Start = RO.scalefac_band.l[i];
	i += gi->region1_count + 1;
	assert(i < 23);
	region2Start = RO.scalefac_band.l[i];

	if (region1Start > bigvalues)
		region1Start = bigvalues;

	if (region2Start > bigvalues)
		region2Start = bigvalues;

	bits += Huffmancodebits(gi->table_select[0], 0, region1Start, ix, xr_sign);
	bits += Huffmancodebits(gi->table_select[1], region1Start, region2Start, ix, xr_sign);
	bits += Huffmancodebits(gi->table_select[2], region2Start, bigvalues, ix, xr_sign);
	return bits;
}

static int writeMainData(gogo_thread_data * tl)
{
	int gr, ch, sfb, data_bits, scale_bits, tot_bits = 0;

	if (RO.version == 1) {
		/* MPEG 1 */
		for (gr = 0; gr < 2; gr++) {
			for (ch = 0; ch < RO.channels_out; ch++) {
				gr_info *gi = &tl->l3_side.tt[gr][ch];

				int slen1 = slen1_tab[gi->scalefac_compress];
				int slen2 = slen2_tab[gi->scalefac_compress];
				data_bits = 0;
				scale_bits = 0;

				if (gi->block_type == SHORT_TYPE) {
					for (sfb = 0; sfb < SBPSY_s; sfb++) {
						int slen = sfb < 6 ? slen1 : slen2;

						assert(tl->scalefac[gr][ch].s[sfb][0] >= 0);
						assert(tl->scalefac[gr][ch].s[sfb][1] >= 0);
						assert(tl->scalefac[gr][ch].s[sfb][2] >= 0);

						putbits2(tl->scalefac[gr][ch].s[sfb][0], slen);
						putbits2(tl->scalefac[gr][ch].s[sfb][1], slen);
						putbits2(tl->scalefac[gr][ch].s[sfb][2], slen);
						scale_bits += 3 * slen;
					}
					data_bits += ShortHuffmancodebits(tl->l3_enc[gr][ch], gi, tl->xr_sign[gr][ch]);
				} else {
					int i;
					for (i = 0; i < sizeof(scfsi_band) / sizeof(int) - 1; i++) {
						if (gr != 0 && tl->l3_side.scfsi[ch][i])
							continue;

						for (sfb = scfsi_band[i]; sfb < scfsi_band[i + 1]; sfb++) {

							assert(tl->scalefac[gr][ch].l[sfb] >= 0);
							putbits2(tl->scalefac[gr][ch].l[sfb], sfb < 11 ? slen1 : slen2);
							scale_bits += sfb < 11 ? slen1 : slen2;
						}
					}
					data_bits += LongHuffmancodebits(tl->l3_enc[gr][ch], gi, tl->xr_sign[gr][ch]);
				}
				data_bits += huffman_coder_count1(tl->l3_enc[gr][ch], gi, tl->xr_sign[gr][ch]);
				/* does bitcount in quantize.c agree with actual bit count? */
				assert(data_bits == gi->part2_3_length - gi->part2_length);
				assert(scale_bits == gi->part2_length);
				tot_bits += scale_bits + data_bits;

			}					/* for ch */
		}						/* for gr */
	} else {
		/* MPEG 2 */
		gr = 0;
		for (ch = 0; ch < RO.channels_out; ch++) {
			gr_info *gi = &tl->l3_side.tt[gr][ch];

			int i, sfb_partition;
			assert(gi->sfb_partition_table);
			data_bits = 0;
			scale_bits = 0;

			sfb = 0;
			sfb_partition = 0;
			if (gi->block_type == SHORT_TYPE) {
				for (; sfb_partition < 4; sfb_partition++) {
					int sfbs = gi->sfb_partition_table[sfb_partition] / 3;
					int slen = gi->slen[sfb_partition];
					for (i = 0; i < sfbs; i++, sfb++) {
						putbits2(Max(tl->scalefac[gr][ch].s[sfb][0], 0U), slen);
						putbits2(Max(tl->scalefac[gr][ch].s[sfb][1], 0U), slen);
						putbits2(Max(tl->scalefac[gr][ch].s[sfb][2], 0U), slen);
						scale_bits += 3 * slen;
					}
				}
				data_bits += ShortHuffmancodebits(tl->l3_enc[gr][ch], gi, tl->xr_sign[gr][ch]);
			} else {
				for (; sfb_partition < 4; sfb_partition++) {
					int sfbs = gi->sfb_partition_table[sfb_partition];
					int slen = gi->slen[sfb_partition];
					for (i = 0; i < sfbs; i++, sfb++) {
						putbits2(Max(tl->scalefac[gr][ch].l[sfb], 0U), slen);
						scale_bits += slen;
					}
				}
				data_bits += LongHuffmancodebits(tl->l3_enc[gr][ch], gi, tl->xr_sign[gr][ch]);
			}
			data_bits += huffman_coder_count1(tl->l3_enc[gr][ch], gi, tl->xr_sign[gr][ch]);

			/* does bitcount in quantize.c agree with actual bit count? */
			assert(data_bits == gi->part2_3_length - gi->part2_length);
			assert(scale_bits == gi->part2_length);
			tot_bits += scale_bits + data_bits;
		} /* for ch */
	} /* for gf */
	return tot_bits;
}

void flush_bitstream(gogo_thread_data * tl)
{
	int flushbits, remaining_headers;
	int bitsPerFrame, mean_bits;
	int last_ptr, first_ptr;

	first_ptr = RW.w_ptr;		/* first header to add to bitstream */
	last_ptr = RW.h_ptr - 1;	/* last header to add to bitstream */
	if (last_ptr == -1)
		last_ptr = MAX_HEADER_BUF - 1;

	/* add this many bits to bitstream so we can flush all headers */
	flushbits = RW.header[last_ptr].write_timing - RW.bs.totbit;

	if (flushbits >= 0) {
		/* if flushbits >= 0, some headers have not yet been written */
		/* reduce flushbits by the size of the headers */
		remaining_headers = 1 + last_ptr - first_ptr;
		if (last_ptr < first_ptr)
			remaining_headers = 1 + last_ptr - first_ptr + MAX_HEADER_BUF;
		flushbits -= remaining_headers * 8 * RO.sideinfo_len;
	}


	/* finally, add some bits so that the last frame is complete
	 * these bits are not necessary to decode the last frame, but
	 * some decoders will ignore last frame if these bits are missing 
	 */
	RO.getframebits(tl, &bitsPerFrame, &mean_bits);
	flushbits += bitsPerFrame;
	assert(flushbits >= 0);
	if (flushbits) {
		drain_into_ancillary(flushbits);
	}

	assert(RW.header[last_ptr].write_timing + bitsPerFrame == RW.bs.totbit);
}

void add_dummy_byte(unsigned char val)
{
	int i;
	putbits_noheaders(val, 8);
	for (i = 0; i < MAX_HEADER_BUF; ++i) {
		RW.header[i].write_timing += 8;
	}
}

/*
  format_bitstream()

  This is called after a frame of audio has been quantized and coded.
  It will write the encoded audio to the bitstream. Note that
  from a layer3 encoder's perspective the bit stream is primarily
  a series of main_data() blocks, with header and side information
  inserted at the proper locations to maintain framing. (See Figure A.7
  in the IS).
  */
int format_bitstream(gogo_thread_data * tl, int bitsPerFrame)
{
	int bits;
	III_side_info_t *l3_side;
	l3_side = &tl->l3_side;

	drain_into_ancillary(l3_side->resvDrain_pre);

	encodeSideInfo2(tl, bitsPerFrame);
	bits = 8 * RO.sideinfo_len;
	bits += writeMainData(tl);
	drain_into_ancillary(l3_side->resvDrain_post);
	bits += l3_side->resvDrain_post;

	RW.main_data_begin += (bitsPerFrame - bits) / 8;
	assert(RW.main_data_begin * 8 == tl->ResvSize);
	assert(RW.bs.totbit % 8 == 0);

	if (RW.bs.totbit > 1000000000) {
		/* to avoid totbit overflow, (at 8h encoding at 128kbs) lets reset bit counter */
		int i;
		for (i = 0; i < MAX_HEADER_BUF; ++i) {
			RW.header[i].write_timing -= RW.bs.totbit;
		}
		RW.bs.totbit = 0;
	}
	return 0;
}

int put_mp3(Bit_stream_struc * bs)
{
	int minimum;
	minimum = bs->buf_byte_idx + 1;
	if (minimum == 0)
		return 0;
	assert(minimum <= LAME_MAXMP3BUFFER);
	bs->buf_byte_idx = -1;
	bs->buf_bit_idx = 0;
	if (writeData(bs->buf, minimum) != minimum)
		return -1;
	return minimum;
}

void init_bit_stream_w(void)
{
	static unsigned char s_buf[LAME_MAXMP3BUFFER];
	RW.bs.buf = s_buf;
	RW.bs.buf_size = LAME_MAXMP3BUFFER;

	RW.h_ptr = RW.w_ptr = 0;
	RW.header[RW.h_ptr].write_timing = 0;
	RW.bs.buf_byte_idx = -1;
	RW.bs.buf_bit_idx = 0;
	RW.bs.totbit = 0;
}

/* end of bitstream.c */
