/*
 *	Xing VBR tagging for LAME.
 *
 *	Copyright (c) 1999 A.L. Faber
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include <assert.h>

#include "config.h"
#include "global.h"

#include "vbrtag.h"
#include "bitstream.h"
#include "util.h"
#include "lame.h"

#ifdef _DEBUG
/*  #define DEBUG_VBRTAG */
#endif

/*
//    4 bytes for Header Tag
//    4 bytes for Header Flags
//  100 bytes for entry (NUMTOCENTRIES)
//    4 bytes for FRAME SIZE
//    4 bytes for STREAM_SIZE
//    4 bytes for VBR SCALE. a VBR quality indicator: 0=best 100=worst
//   20 bytes for LAME tag.  for example, "LAME3.12 (beta 6)"
// ___________
//  140 bytes
*/
#define VBRHEADERSIZE (NUMTOCENTRIES+4+4+4+4+4)

/* the size of the Xing header (MPEG1 and MPEG2) in kbps */
#define XING_BITRATE1 128
#define XING_BITRATE2  64
#define XING_BITRATE25 32



static const char VBRTag[]={"Xing"};
static const int SizeOfEmptyFrame[2][2]=
{
	{17,9},
	{32,17},
};


/***********************************************************************
 *  Robert Hegemann 2001-01-17
 ***********************************************************************/

static void addVbr(VBR_seek_info_t * v, int bitrate)
{
    int i;

    v->sum += bitrate;
    v->seen ++;
    
    if (v->seen < v->want) {
        return;
    }

    if (v->pos < v->size) {
        v->bag[v->pos] = v->sum;
        v->pos ++;
        v->seen = 0;
    }
    if (v->pos == v->size) {
        for (i = 1; i < v->size; i += 2) {
            v->bag[i/2] = v->bag[i]; 
        }
        v->want *= 2;
        v->pos  /= 2;
    }
}

static void Xing_seek_table(VBR_seek_info_t * v, unsigned char *t)
{
    int i, index;
    int seek_point;
    
    if (v->pos <= 0)
        return;
        
    for (i = 1; i < NUMTOCENTRIES; ++i) {
        float j = i/(float)NUMTOCENTRIES, act, sum;
        index = (int)(floor(j * v->pos));
        if (index > v->pos-1)
            index = v->pos-1;
        act = v->bag[index];
        sum = v->sum;
        seek_point = (int)(256. * act / sum);
        if (seek_point > 255)
            seek_point = 255;
        t[i] = seek_point;
    }
}

#if 0
void print_seeking(unsigned char *t)
{
    int i;
    
    printf("seeking table ");
    for (i = 0; i < NUMTOCENTRIES; ++i) {
        printf(" %d ", t[i]);
    }
    printf("\n");
}
#endif



/****************************************************************************
 * AddVbrFrame: Add VBR entry, used to fill the VBR the TOC entries
 * Paramters:
 *	nStreamPos: how many bytes did we write to the bitstream so far
 *				(in Bytes NOT Bits)
 ****************************************************************************
*/
void
AddVbrFrame(int bitrate_index)
{
    int kbps = bitrate_table[RO.version][bitrate_index];
    
    addVbr(&RW.VBR_seek_table, kbps);
//    RW.nVbrNumFrames++;
}

#ifdef DEBUG_VBRTAG
/*-------------------------------------------------------------*/
static int ExtractI4(unsigned char *buf)
{
	int x;
	/* big endian extract */
	x = buf[0];
	x <<= 8;
	x |= buf[1];
	x <<= 8;
	x |= buf[2];
	x <<= 8;
	x |= buf[3];
	return x;
}
#endif

void CreateI4(unsigned char *buf, int nValue)
{
        /* big endian create */
	buf[0]=(nValue>>24)&0xff;
	buf[1]=(nValue>>16)&0xff;
	buf[2]=(nValue>> 8)&0xff;
	buf[3]=(nValue    )&0xff;
}


#ifdef DEBUG_VBRTAG
/*-------------------------------------------------------------*/
/* Same as GetVbrTag below, but only checks for the Xing tag.
   requires buf to contain only 40 bytes */
/*-------------------------------------------------------------*/
int CheckVbrTag(unsigned char *buf)
{
	int			h_id, h_mode, h_sr_index;

	/* get selected MPEG header data */
	h_id       = (buf[1] >> 3) & 1;
	h_sr_index = (buf[2] >> 2) & 3;
	h_mode     = (buf[3] >> 6) & 3;

	/*  determine offset of header */
	if( h_id )
	{
                /* mpeg1 */
		if( h_mode != 3 )	buf+=(32+4);
		else				buf+=(17+4);
	}
	else
	{
                /* mpeg2 */
		if( h_mode != 3 ) buf+=(17+4);
		else              buf+=(9+4);
	}

	if( buf[0] != VBRTag[0] ) return 0;    /* fail */
	if( buf[1] != VBRTag[1] ) return 0;    /* header not found*/
	if( buf[2] != VBRTag[2] ) return 0;
	if( buf[3] != VBRTag[3] ) return 0;
	return 1;
}
#endif

#ifdef DEBUG_VBRTAG
int GetVbrTag(VBRTAGDATA *pTagData,  unsigned char *buf)
{
	int			i, head_flags;
	int			h_bitrate,h_id, h_mode, h_sr_index;

	/* get Vbr header data */
	pTagData->flags = 0;

	/* get selected MPEG header data */
	h_id       = (buf[1] >> 3) & 1;
	h_sr_index = (buf[2] >> 2) & 3;
	h_mode     = (buf[3] >> 6) & 3;
        h_bitrate  = ((buf[2]>>4)&0xf);
	h_bitrate = bitrate_table[h_id][h_bitrate];


	/*  determine offset of header */
	if( h_id )
	{
                /* mpeg1 */
		if( h_mode != 3 )	buf+=(32+4);
		else				buf+=(17+4);
	}
	else
	{
                /* mpeg2 */
		if( h_mode != 3 ) buf+=(17+4);
		else              buf+=(9+4);
	}

	if( buf[0] != VBRTag[0] ) return 0;    /* fail */
	if( buf[1] != VBRTag[1] ) return 0;    /* header not found*/
	if( buf[2] != VBRTag[2] ) return 0;
	if( buf[3] != VBRTag[3] ) return 0;

	buf+=4;

	pTagData->h_id = h_id;
	pTagData->samprate = samplerate_table[h_id][h_sr_index];

	if( h_id == 0 )
		pTagData->samprate >>= 1;

	head_flags = pTagData->flags = ExtractI4(buf); buf+=4;      /* get flags */

	if( head_flags & FRAMES_FLAG )
	{
		pTagData->frames   = ExtractI4(buf); buf+=4;
	}

	if( head_flags & BYTES_FLAG )
	{
		pTagData->bytes = ExtractI4(buf); buf+=4;
	}

	if( head_flags & TOC_FLAG )
	{
		if( pTagData->toc != NULL )
		{
			for(i=0;i<NUMTOCENTRIES;i++)
				pTagData->toc[i] = buf[i];
		}
		buf+=NUMTOCENTRIES;
	}

	pTagData->vbr_scale = -1;

	if( head_flags & VBR_SCALE_FLAG )
	{
		pTagData->vbr_scale = ExtractI4(buf); buf+=4;
	}

	pTagData->headersize = 
	  ((h_id+1)*72000*h_bitrate) / pTagData->samprate;


#ifdef DEBUG_VBRTAG
	RO.printf("\n\n********************* VBR TAG INFO *****************\n");
	RO.printf("tag         :%s\n",VBRTag);
	RO.printf("head_flags  :%d\n",head_flags);
	RO.printf("bytes       :%d\n",pTagData->bytes);
	RO.printf("frames      :%d\n",pTagData->frames);
	RO.printf("VBR Scale   :%d\n",pTagData->vbr_scale);
	RO.printf("toc:\n");
	if( pTagData->toc != NULL )
	{
		for(i=0;i<NUMTOCENTRIES;i++)
		{
			if( (i%10) == 0 ) RO.printf("\n");
			RO.printf(" %3d", (int)(pTagData->toc[i]));
		}
	}
	RO.printf("\n***************** END OF VBR TAG INFO ***************\n");
#endif
	return 1;       /* success */
}
#endif


/****************************************************************************
 * InitVbrTag: Initializes the header, and write empty frame to stream
 * Paramters:
 *				fpStream: pointer to output file stream
 *				nMode	: Channel Mode: 0=STEREO 1=JS 2=DS 3=MONO
 ****************************************************************************
*/
int InitVbrTag(void)
{
	int nMode;
#define MAXFRAMESIZE 576
	//	u_char pbtStreamBuffer[MAXFRAMESIZE];
	nMode = RO.mode;

	/* Clear Frame position array variables */
//	RW.nVbrNumFrames=0;

	/* Reserve the proper amount of bytes */
	if (nMode == MONO) {
		BE.nZeroStreamSize=SizeOfEmptyFrame[RO.version][1]+4;
	} else {
		BE.nZeroStreamSize=SizeOfEmptyFrame[RO.version][0]+4;
	}

	/*
	// Xing VBR pretends to be a 48kbs layer III frame.  (at 44.1kHz).
        // (at 48kHz they use 56kbs since 48kbs frame not big enough for
        // table of contents)
	// let's always embed Xing header inside a 64kbs layer III frame.
	// this gives us enough room for a LAME version string too.
	// size determined by sampling frequency (MPEG1)
	// 32kHz:    216 bytes@48kbs    288bytes@ 64kbs
	// 44.1kHz:  156 bytes          208bytes@64kbs     (+1 if padding = 1)
	// 48kHz:    144 bytes          192
	//
	// MPEG 2 values are the same since the framesize and samplerate
        // are each reduced by a factor of 2.
	*/
	{
	int i,bitrate;
	if (1==RO.version) {
	  bitrate = XING_BITRATE1;
	} else {
	  if (RO.out_samplerate < 16000 )
	    bitrate = XING_BITRATE25;
	  else
	    bitrate = XING_BITRATE2;
	}
	BE.TotalFrameSize= ((RO.version+1)*72000*bitrate) / RO.out_samplerate;
	/* extra 20 bytes for LAME & version string */
	assert(BE.TotalFrameSize >= (BE.nZeroStreamSize+VBRHEADERSIZE) + 20);
	assert(BE.TotalFrameSize <= MAXFRAMESIZE );

	for (i=0; i<BE.TotalFrameSize; ++i)
	  add_dummy_byte(0);
	}

        RW.VBR_seek_table.sum  = 0;
        RW.VBR_seek_table.seen = 0;
        RW.VBR_seek_table.want = 1;
        RW.VBR_seek_table.pos  = 0;
        RW.VBR_seek_table.size = 400;

	/* Success */
	return 0;
}


/****************************************************************************
 * CreateVbrTag: Create VBR tag 
 * Paramters:
 ****************************************************************************
*/
int CreateVbrTag(void **pTagData,size_t *pSizeResult, int nVbrScale, const size_t sizeWrite)
{
	long lFileSize;
	int nStreamIndex;
	char bbyte;
	u_char		btToc[NUMTOCENTRIES];
	u_char pbtStreamBuffer[MAXFRAMESIZE];
	char str1[80];
//        unsigned char id3v2Header[10];
//        const size_t id3v2TagSize = 0;
	gogo_thread_data *tl = RO.tl; 
		
//		FILE *fpStream = NULL;
	if (RW.VBR_seek_table.pos <= 0)
		return -1;


	/* Clear stream buffer */
	memset(pbtStreamBuffer,0x00,sizeof(pbtStreamBuffer));

	/* Get file size */
	lFileSize = sizeWrite;

	/* Abort if file has zero length. Yes, it can happen :) */
	if (lFileSize==0)
		return -1;

	/* the default VBR header. 48 kbps layer III, no padding, no crc */
	/* but sampling freq, mode andy copyright/copy protection taken */
	/* from first valid frame */
	pbtStreamBuffer[0] = (u_char) 0xff;
	pbtStreamBuffer[1] = (u_char) ( RO.out_samplerate < 16000 ) ? 0xE0 : 0xF0;
	pbtStreamBuffer[1] |= 
					(RO.version << 3) | 
					((4 - 3 /* Layer3 */) << 1) | 
					1; /* Error Protection */;

	{
		int bitrate;
		if (1==RO.version) {
			bitrate = XING_BITRATE1;
		} else {
			if (RO.out_samplerate < 16000 )
				bitrate = XING_BITRATE25;
			else
				bitrate = XING_BITRATE2;
		}
		bbyte = 16*BitrateIndex(bitrate,RO.version);
	}

	pbtStreamBuffer[2]= bbyte | 
			(RO.samplerate_index << 2) |
			( tl->padding << 1) |
			0;								/* 0 extension */

	pbtStreamBuffer[3]=
				(RO.mode << 6)		|
				(tl->mode_ext << 4)	|
				(COPYRIGHT << 3 ) |
				(ORIGINAL << 2 )	|
				(RO.emphasis << 0);
				
	/* Clear all TOC entries */
	memset(btToc,0,sizeof(btToc));

        Xing_seek_table (&RW.VBR_seek_table, btToc);
        /* print_seeking (btToc); */

	/* Start writing the tag after the zero frame */
	nStreamIndex=BE.nZeroStreamSize;

	/* Put Vbr tag */
	pbtStreamBuffer[nStreamIndex++]=VBRTag[0];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[1];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[2];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[3];

	/* Put header flags */
	CreateI4(&pbtStreamBuffer[nStreamIndex],FRAMES_FLAG+BYTES_FLAG+TOC_FLAG+VBR_SCALE_FLAG);
	nStreamIndex+=4;

	/* Put Total Number of frames */
	CreateI4(&pbtStreamBuffer[nStreamIndex],RW.frameNum );
	nStreamIndex+=4;

	/* Put Total file size */
	CreateI4(&pbtStreamBuffer[nStreamIndex],(int)lFileSize);
	nStreamIndex+=4;

	/* Put TOC */
	memcpy(&pbtStreamBuffer[nStreamIndex],btToc,sizeof(btToc));
	nStreamIndex+=sizeof(btToc);

	/* Put VBR SCALE */
	CreateI4(&pbtStreamBuffer[nStreamIndex],nVbrScale);
	nStreamIndex+=4;

	/* Put GOGO ID */
	strcpy(str1, "GOGO");
    strncpy((char *)&pbtStreamBuffer[nStreamIndex], str1, 20 );
    nStreamIndex += 20;

#if 1
	{
		static char s_buf[2*9*384];
		assert( BE.TotalFrameSize <= sizeof(s_buf) );
		memcpy( s_buf, pbtStreamBuffer, BE.TotalFrameSize );

		*pSizeResult	= BE.TotalFrameSize;
		*pTagData		= s_buf;
	}
#else
	{
		static char s_buf[2*9*384];
		void	*pAllocData = malloc( BE.TotalFrameSize );
		if( pAllocData == NULL ) return -1;
		memcpy( pAllocData, pbtStreamBuffer, BE.TotalFrameSize );

		*pSizeResult	= BE.TotalFrameSize;
		*pTagData		= pAllocData;
	}
#endif
	return 0;       /* success */
}
